"""Classes used for converting molecules and rewrites into graphs compatible with `GNNEncoder`."""

from __future__ import annotations

import hashlib
from abc import ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from typing import Any, Iterable, Optional, Union

import numpy as np
from numpy.typing import NDArray
from rdkit import Chem
from syntheseus.interface.molecule import Molecule

from retrochimera.chem.rewrite import Rewrite
from retrochimera.dgllife.featurizers import (
    BaseAtomFeaturizer,
    BaseBondFeaturizer,
    CanonicalBondFeaturizer,
    ConcatFeaturizer,
    WeaveAtomFeaturizer,
    atom_chiral_tag_one_hot,
    atom_degree_one_hot,
    atom_formal_charge,
    atom_hybridization_one_hot,
    atom_implicit_valence_one_hot,
    atom_is_aromatic,
    atom_num_radical_electrons,
    atom_total_num_H_one_hot,
    atom_type_one_hot,
)

AtomFeaturizer = Union[BaseAtomFeaturizer, WeaveAtomFeaturizer]


@dataclass
class RawGraph:
    """Class to store a featurized graph.

    This is similar to `torch_geometric.data.Data` but uses numpy arrays instead of torch tensors.
    """

    atom_features: NDArray
    bond_features: NDArray
    bonds: NDArray

    atom_categorical_features: Optional[NDArray]


@dataclass
class RawRewriteGraph(RawGraph):
    """Featurized graph which has its nodes divided into two parts: 'left' and 'right'.

    Technically this is not a bipartite graph, as nodes within each group can be connected.
    """

    node_in_lhs: NDArray


class Featurizer(metaclass=ABCMeta):
    @abstractproperty
    def num_atom_features(self) -> int:
        pass

    @abstractproperty
    def num_bond_features(self) -> int:
        pass

    @property
    def num_atom_categorical_features(self) -> list[int]:
        return []


class MoleculeFeaturizer(Featurizer):
    @abstractmethod
    def __call__(self, mol: Molecule) -> RawGraph:
        pass


class RewriteFeaturizer(Featurizer):
    @abstractmethod
    def __call__(self, rewrite: Rewrite) -> RawGraph:
        pass

    @staticmethod
    def prepare_kwargs(rewrites: Iterable[Rewrite]) -> dict[str, Any]:
        return {}


def featurize_with_dgl_life(
    mol: Chem.Mol, atom_featurizer: AtomFeaturizer, bond_featurizer: BaseBondFeaturizer
) -> RawGraph:
    """Calls the DGL-Life atom/bond featurizers on a given `Chem.Mol` object.

    Importantly, this function does not perform any sanitization on `mol`; it does not have to be a
    valid molecule as long as it is accepted by the provided featurizers. `DGLLifeRewriteFeaturizer`
    exploits this, passing a `mol` constructed from a SMARTS string.
    """
    atom_features = atom_featurizer(mol)["h"]

    if len(mol.GetBonds()) > 0:
        bonds_list = []
        for bond in mol.GetBonds():
            bonds_list.append((bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()))
            bonds_list.append((bond.GetEndAtomIdx(), bond.GetBeginAtomIdx()))

        bonds = np.asarray(bonds_list)
        bond_features = bond_featurizer(mol)["e"]
    else:
        # The `dgllife` featurizer returns an empty dictionary if there are no bonds, which
        # would lead the above to fail; we handle that case separately.
        bonds = np.zeros((0, 2))
        bond_features = np.zeros((0, bond_featurizer.feat_size()))

    return RawGraph(
        atom_features=atom_features,
        bond_features=bond_features,
        bonds=bonds,
        atom_categorical_features=None,
    )


class DGLLifeMoleculeFeaturizer(MoleculeFeaturizer):
    """A molecule featurizer wrapping around classes provided by DGL-Life."""

    @staticmethod
    def get_default(atom_types: Optional[list[str]] = None) -> DGLLifeMoleculeFeaturizer:
        """Creates a default featurizer.

        Args:
            atom_types: List of atom types to consider. If set to `None`, will use the same
                extensive list of atom types as used in the LocalRetro model.
        """
        # fmt: off
        DEFAULT_ATOM_TYPES = [
            "C", "N", "O", "S", "F", "Si", "P", "Cl", "Br", "Mg", "Na", "Ca", "Fe",
            "As", "Al", "I", "B", "V", "K", "Tl", "Yb", "Sb", "Sn", "Ag", "Pd", "Co",
            "Se", "Ti", "Zn", "H", "Li", "Ge", "Cu", "Au", "Ni", "Cd", "In", "Mn", "Zr",
            "Cr", "Pt", "Hg", "Pb", "W", "Ru", "Nb", "Re", "Te", "Rh", "Tc", "Ba", "Bi",
            "Hf", "Mo", "U", "Sm", "Os", "Ir", "Ce", "Gd", "Ga", "Cs", "Ta"
        ]
        # fmt: on

        atom_featurizer = WeaveAtomFeaturizer(atom_types=atom_types or DEFAULT_ATOM_TYPES)
        bond_featurizer = CanonicalBondFeaturizer(self_loop=False)

        return DGLLifeMoleculeFeaturizer(atom_featurizer, bond_featurizer)

    def __init__(
        self, atom_featurizer: AtomFeaturizer, bond_featurizer: BaseBondFeaturizer
    ) -> None:
        self.atom_featurizer = atom_featurizer
        self.bond_featurizer = bond_featurizer

    @staticmethod
    def _featurize_single(
        mol: Molecule, atom_featurizer: AtomFeaturizer, bond_featurizer: BaseBondFeaturizer
    ) -> RawGraph:
        return featurize_with_dgl_life(
            mol=mol.rdkit_mol,
            atom_featurizer=atom_featurizer,
            bond_featurizer=bond_featurizer,
        )

    def __call__(self, mol: Molecule) -> RawGraph:
        return DGLLifeMoleculeFeaturizer._featurize_single(
            mol,
            atom_featurizer=self.atom_featurizer,
            bond_featurizer=self.bond_featurizer,
        )

    @property
    def num_atom_features(self) -> int:
        return self.atom_featurizer.feat_size()

    @property
    def num_bond_features(self) -> int:
        return self.bond_featurizer.feat_size()


class DGLLifeRewriteFeaturizer(RewriteFeaturizer):
    """A rewrite featurizer wrapping around classes provided by DGL-Life.

    Even though the featurizers provided by DGL-Life are designed for molecules (i.e. SMILES), some
    of them (generally, the more basic ones) also work on rewrites (i.e. SMARTS). This class
    exploits that to featurize both sides of the rewrite, which are glued using the atom mapping.
    """

    @staticmethod
    def get_default(atom_smarts_vocab: Optional[list[str]]) -> DGLLifeRewriteFeaturizer:
        """Creates a default featurizer.

        Args:
            atom_smarts_vocab: Vocabulary of supported atom SMARTS, used for one-hot encoding. If
                set to `None`, the atom SMARTS will instead be featurized with random vectors of a
                fixed size. This option is useful for hashing, where the embeddings do not have to
                be informative or easy to separate.
        """
        # Same as `CanonicalAtomFeaturizer` but with `atom_chiral_tag_one_hot` added.
        atom_featurizer = BaseAtomFeaturizer(
            featurizer_funcs={
                "h": ConcatFeaturizer(
                    [
                        atom_chiral_tag_one_hot,
                        atom_degree_one_hot,
                        atom_formal_charge,
                        atom_hybridization_one_hot,
                        atom_implicit_valence_one_hot,
                        atom_is_aromatic,
                        atom_num_radical_electrons,
                        atom_total_num_H_one_hot,
                        atom_type_one_hot,
                    ]
                )
            }
        )
        bond_featurizer = CanonicalBondFeaturizer(self_loop=False)

        return DGLLifeRewriteFeaturizer(atom_featurizer, bond_featurizer, atom_smarts_vocab)

    def __init__(
        self,
        atom_featurizer: AtomFeaturizer,
        bond_featurizer: BaseBondFeaturizer,
        atom_smarts_vocab: Optional[list[str]],
    ) -> None:
        self.atom_featurizer = atom_featurizer
        self.bond_featurizer = bond_featurizer

        self._atom_smarts_to_id: Optional[dict[str, int]]
        if atom_smarts_vocab is not None:
            self._atom_smarts_to_id = {
                atom_smarts: id for id, atom_smarts in enumerate(atom_smarts_vocab)
            }
        else:
            self._atom_smarts_to_id = None

    @staticmethod
    def _smarts_to_mol(smarts: str) -> Chem.Mol:
        mol = Chem.MolFromSmarts(smarts)
        mol.UpdatePropertyCache(strict=False)
        return mol

    @staticmethod
    def _get_atom_smarts(mol: Chem.Mol) -> list[str]:
        return [atom.GetSmarts() for atom in mol.GetAtoms()]

    @staticmethod
    def prepare_kwargs(rewrites: Iterable[Rewrite]) -> dict[str, Any]:
        atom_smarts_vocab = set()

        for rewrite in rewrites:
            for smarts in [rewrite.lhs, rewrite.rhs]:
                mol = DGLLifeRewriteFeaturizer._smarts_to_mol(smarts)
                atom_smarts_vocab |= set(DGLLifeRewriteFeaturizer._get_atom_smarts(mol))

        return {"atom_smarts_vocab": sorted(list(atom_smarts_vocab))}

    @staticmethod
    def _embed_atom_smarts(smarts: str, atom_smarts_to_id: Optional[dict[str, int]]) -> int:
        if atom_smarts_to_id is not None:
            return atom_smarts_to_id[smarts]
        else:
            return int(hashlib.shake_256(smarts.encode()).hexdigest(4), 16)

    @staticmethod
    def _featurize_single(
        rewrite: Rewrite,
        atom_featurizer: AtomFeaturizer,
        bond_featurizer: BaseBondFeaturizer,
        atom_smarts_to_id: Optional[dict[str, int]],
    ) -> RawRewriteGraph:
        graphs = []
        atom_smarts = []

        # First, featurize each side separately using DGL-Life.
        for smarts in [rewrite.lhs, rewrite.rhs]:
            mol = DGLLifeRewriteFeaturizer._smarts_to_mol(smarts)
            atom_smarts += DGLLifeRewriteFeaturizer._get_atom_smarts(mol)
            graphs.append(
                featurize_with_dgl_life(
                    mol=mol, atom_featurizer=atom_featurizer, bond_featurizer=bond_featurizer
                )
            )

        # TODO(krmaziar): Consider handling bond SMARTS in a similar way.
        atom_smarts_embeddings = np.stack(
            [
                DGLLifeRewriteFeaturizer._embed_atom_smarts(
                    smarts, atom_smarts_to_id=atom_smarts_to_id
                )
                for smarts in atom_smarts
            ]
        )

        [num_atoms_lhs, num_atoms_rhs] = [len(graph.atom_features) for graph in graphs]

        # Build an extra binary feature to differentiate left from right.
        which_side_feature = np.concatenate(
            [np.ones((num_atoms_lhs, 1)), np.zeros((num_atoms_rhs, 1))], axis=0
        )

        # Grab bonds from both sides, shift the atom ids for the rhs.
        [bonds_lhs, bonds_rhs] = [graph.bonds for graph in graphs]

        combined_graph = RawRewriteGraph(
            atom_features=np.concatenate([graph.atom_features for graph in graphs], axis=0),
            bond_features=np.concatenate([graph.bond_features for graph in graphs], axis=0),
            bonds=np.concatenate([bonds_lhs, bonds_rhs + num_atoms_lhs], axis=0),
            atom_categorical_features=atom_smarts_embeddings[:, np.newaxis],
            # Useful downstream if we only care about the lhs representations:
            node_in_lhs=np.squeeze(which_side_feature, axis=-1),
        )

        # Concatenate the lhs-vs-rhs feature into the atom features.
        combined_graph.atom_features = np.concatenate(
            [combined_graph.atom_features, which_side_feature], axis=1
        )

        num_bonds_original = len(combined_graph.bonds)

        # Prepare the atom mapping edges that will connect the two components.
        atom_mapping_bonds_list = [
            (atom_idx_lhs, atom_idx_rhs + num_atoms_lhs)
            for (atom_idx_lhs, atom_idx_rhs) in rewrite.mapping
        ]

        # Include the new edges in both directions.
        atom_mapping_bonds_list += [
            (idx_rhs, idx_lhs) for (idx_lhs, idx_rhs) in atom_mapping_bonds_list
        ]

        if atom_mapping_bonds_list:
            atom_mapping_bonds = np.asarray(atom_mapping_bonds_list)
        else:
            atom_mapping_bonds = np.zeros((0, 2))

        combined_graph.bonds = np.concatenate([combined_graph.bonds, atom_mapping_bonds], axis=0)
        combined_graph.bond_features = np.concatenate(
            [
                combined_graph.bond_features,
                np.zeros((len(atom_mapping_bonds_list), combined_graph.bond_features.shape[1])),
            ],
            axis=0,
        )

        # Add an extra binary feature to differentiate real bonds from the atom mapping.
        is_atom_mapping_feature = np.concatenate(
            [np.zeros((num_bonds_original, 1)), np.ones((len(atom_mapping_bonds_list), 1))], axis=0
        )
        combined_graph.bond_features = np.concatenate(
            [combined_graph.bond_features, is_atom_mapping_feature], axis=1
        )

        return combined_graph

    def __call__(self, rewrite: Rewrite) -> RawGraph:
        return DGLLifeRewriteFeaturizer._featurize_single(
            rewrite,
            atom_featurizer=self.atom_featurizer,
            bond_featurizer=self.bond_featurizer,
            atom_smarts_to_id=self._atom_smarts_to_id,
        )

    @property
    def num_atom_features(self) -> int:
        return self.atom_featurizer.feat_size() + 1

    @property
    def num_bond_features(self) -> int:
        return self.bond_featurizer.feat_size() + 1

    @property
    def num_atom_categorical_features(self) -> list[int]:
        assert self._atom_smarts_to_id is not None
        return [len(self._atom_smarts_to_id)]



import pytest
from rdkit import Chem
from rdkit.Chem import AllChem

from retrochimera.chem.utils import get_atom_mapping, normalize_atom_mapping


def test_get_atom_mapping() -> None:
    mol = Chem.MolFromSmarts("[C:1]-[C&H0&+0&D3:5](=[O&H0&D1:4])-O.[C:2]-[N&H2&+0&D1:3]")
    atom_mapping = get_atom_mapping(mol)

    assert atom_mapping == {1: 0, 2: 4, 3: 5, 4: 2, 5: 1}


@pytest.mark.parametrize(
    "smarts, smarts_normalized",
    [
        ("[c:10]-[N:2]>>[c:10]-[N:2](=O)-[O&-:5]", "[c:2]-[N:1]>>[c:2]-[N:1](=O)-[O&-]"),
        ("[#6:3]-[#7:28]>>[#8:12]-[#6:3].[#7:28]", "[#6:1]-[#7:2]>>[#8]-[#6:1].[#7:2]"),
    ],
)
def test_normalize_mapping(smarts: str, smarts_normalized: str) -> None:
    template = AllChem.ReactionFromSmarts(smarts)
    normalize_atom_mapping(template)

    assert AllChem.ReactionToSmarts(template) == smarts_normalized



from dataclasses import dataclass

from omegaconf import MISSING
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config

from retrochimera.data.smiles_tokenizer import Tokenizer
from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)


@dataclass
class BuildTokenizerConfig:
    """Config for building a tokenizer for training SMILES-based models."""

    data_dir: str = MISSING  # Directory containing raw data
    output_vocab_path: str = MISSING  # Path for saving the resulting vocabulary file


def main() -> None:
    config: BuildTokenizerConfig = cli_get_config(argv=None, config_cls=BuildTokenizerConfig)
    logger.info(f"Building a tokenizer with the following config: {config}")

    tokenizer = Tokenizer.from_reactions(reaction_data_path=config.data_dir)
    tokenizer.save_vocab(vocab_path=config.output_vocab_path)


if __name__ == "__main__":
    main()



import os

# Needed for cases when reaction data preprocessing does not release the lock on the `*.h5` file.
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"


from retrochimera.inference import (
    BasicTemplateClassificationModel,
    RetroChimeraDeNovoModel,
    RetroChimeraEditModel,
    RetroChimeraModel,
)

__all__ = [
    "BasicTemplateClassificationModel",
    "RetroChimeraDeNovoModel",
    "RetroChimeraEditModel",
    "RetroChimeraModel",
]



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/encoders/transformer.py
Modifications:
1. Removed the `parallel_gpu` argument from the `TransformerEncoderLayer` and `TransformerEncoder` constructors, as Chimera typically uses Pytorch-Lightning for distributed training.
2. Removed the `embeddings` argument from the `TransformerEncoder` constructor, as it is now built in the `retrochimera.models.smiles_transformer.SmilesTransformerModel` class.
3. Refactored the `forward` method of the `TransformerEncoder` class to align with the `retrochimera.models.smiles_transformer.SmilesTransformerModel` class.
"""
import torch
import torch.nn as nn

from retrochimera.opennmt.modules.multi_head_attention import MultiHeadedAttention
from retrochimera.opennmt.modules.position_ffn import ActivationFunction, PositionwiseFeedForward
from retrochimera.opennmt.modules.rmsnorm import RMSNorm


class TransformerEncoderLayer(nn.Module):
    """
    A single layer of the transformer encoder.

    Args:
        d_model (int): the dimension of keys/values/queries in MultiHeadedAttention, also the input size of
        the first-layer of the PositionwiseFeedForward.
        heads (int): the number of head for MultiHeadedAttention.
        d_ff (int): the second-layer of the PositionwiseFeedForward.
        dropout (float): dropout probability(0-1.0) used in any other place besides attention layer.
        attention_dropout (float): dropout probability(0-1.0) used in attention layer.
        max_relative_positions (int): enable relative position encoding: -1 to enable Rotary Embeddings,
        > 0 (e.g., 16, 32) to use maximum distance between inputs in relative positions representations
        relative_positions_buckets (int): number of buckets when using relative position bias.
        pos_ffn_activation_fn (str): activation function choice for PositionwiseFeedForward layer.
        add_qkvbias (bool): whether to add bias to the Key/Value nn.Linear.
        num_kv (int): number of heads for KV when different vs Q (multiquery).
        add_ffnbias (bool): whether to add bias to the FF nn.Linear.
        parallel_residual (bool): Use parallel residual connections in each layer block, as used by the GPT-J and GPT-NeoX models.
        layer_norm (string): type of layer normalization standard/rms.
        norm_eps (float): layer norm epsilon.
        use_ckpting (List): layers for which we checkpoint for backward.
        rotary_interleave (bool): Interleave the head dimensions when rotary embeddings are applied.
        rotary_theta (float): rotary base theta.
        rotary_dim (int): rotary dim when different to dim per head.
    """

    def __init__(
        self,
        d_model: int = 256,
        heads: int = 8,
        d_ff: int = 2048,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        max_relative_positions: int = 0,
        relative_positions_buckets: int = 0,
        pos_ffn_activation_fn: str = ActivationFunction.relu,
        add_qkvbias: bool = False,
        num_kv: int = 0,
        add_ffnbias: bool = True,
        parallel_residual: bool = False,
        layer_norm: str = "standard",
        norm_eps: float = 1e-6,
        use_ckpting: list = [],
        rotary_interleave: bool = True,
        rotary_theta: float = 1e4,
        rotary_dim: int = 0,
    ):
        super(TransformerEncoderLayer, self).__init__()

        self.self_attn = MultiHeadedAttention(
            heads,
            d_model,
            dropout=attention_dropout,
            is_decoder=False,
            max_relative_positions=max_relative_positions,
            relative_positions_buckets=relative_positions_buckets,
            rotary_interleave=rotary_interleave,
            rotary_theta=rotary_theta,
            rotary_dim=rotary_dim,
            attn_type="self",
            add_qkvbias=add_qkvbias,
            num_kv=num_kv,
            use_ckpting=use_ckpting,
        )
        self.feed_forward = PositionwiseFeedForward(
            d_model,
            d_ff,
            dropout,
            pos_ffn_activation_fn,
            add_ffnbias,
            parallel_residual,
            layer_norm,
            norm_eps,
            use_ckpting=use_ckpting,
        )
        self.parallel_residual = parallel_residual
        if layer_norm == "standard":
            self.layer_norm = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms":
            self.layer_norm = RMSNorm(d_model, eps=norm_eps)
        else:
            raise ValueError(f"{layer_norm} layer norm type is not supported")
        self.dropout_p = dropout
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        layer_in: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Args:
            layer_in (FloatTensor): shape (batch_size, src_len, hidden_dim).
            mask (LongTensor): shape (batch_size, 1, src_len).

        Returns:
            layer_out (FloatTensor): shape (batch_size, src_len, hidden_dim).
        """
        norm_layer_in = self.layer_norm(layer_in)
        context, _ = self.self_attn(norm_layer_in, norm_layer_in, norm_layer_in, mask=mask)
        if self.dropout_p > 0:
            context = self.dropout(context)
        if self.parallel_residual:
            # feed_forward applies residual, so we remove and apply residual with un-normed
            layer_out = self.feed_forward(norm_layer_in) - norm_layer_in + layer_in + context
        else:
            layer_out = context + layer_in
            layer_out = self.feed_forward(layer_out)

        return layer_out

    def update_dropout(
        self,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
    ):
        self.self_attn.update_dropout(attention_dropout)
        self.feed_forward.update_dropout(dropout)
        self.dropout.p = dropout


class TransformerEncoder(nn.Module):
    """The Transformer encoder from "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`

    Args:
        num_layers (int): number of encoder layers.
        d_model (int): size of the model.
        heads (int): number of heads.
        d_ff (int): size of the inner FF layer.
        dropout (float): dropout parameters.
        pos_ffn_activation_fn (str): activation function for PositionwiseFeedForward layer.
    """

    def __init__(
        self,
        num_layers: int = 6,
        d_model: int = 256,
        heads: int = 8,
        d_ff: int = 2048,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        max_relative_positions: int = 0,
        relative_positions_buckets: int = 0,
        pos_ffn_activation_fn: str = ActivationFunction.relu,
        add_qkvbias: bool = False,
        num_kv: int = 0,
        add_ffnbias: bool = True,
        parallel_residual: bool = False,
        layer_norm: str = "standard",
        norm_eps: float = 1e-6,
        use_ckpting: list = [],
        rotary_interleave: bool = True,
        rotary_theta: float = 1e4,
        rotary_dim: int = 0,
    ):
        super(TransformerEncoder, self).__init__()

        self.transformer = nn.ModuleList(
            [
                TransformerEncoderLayer(
                    d_model,
                    heads,
                    d_ff,
                    dropout,
                    attention_dropout,
                    max_relative_positions=max_relative_positions,
                    relative_positions_buckets=relative_positions_buckets,
                    pos_ffn_activation_fn=pos_ffn_activation_fn,
                    add_qkvbias=add_qkvbias,
                    num_kv=num_kv,
                    add_ffnbias=add_ffnbias,
                    parallel_residual=parallel_residual,
                    layer_norm=layer_norm,
                    norm_eps=norm_eps,
                    use_ckpting=use_ckpting,
                    rotary_interleave=rotary_interleave,
                    rotary_theta=rotary_theta,
                    rotary_dim=rotary_dim,
                )
                for i in range(num_layers)
            ]
        )
        if layer_norm == "standard":
            self.layer_norm = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms":
            self.layer_norm = RMSNorm(d_model, eps=norm_eps)
        else:
            raise ValueError(f"{layer_norm} layer norm type is not supported")

    def forward(self, src: torch.Tensor, src_key_padding_mask: torch.Tensor):
        """
        Args:
            src (torch.Tensor) - the sequence to the encoder (required). (batch_size, padded_src_len, hidden_dim).
            src_key_padding_mask (torch.Tensor) - the mask for the src keys per batch (required). (batch_size, padded_src_len).

        Returns:
            enc_out (torch.Tensor) - the output of the encoder. (batch_size, padded_src_len, hidden_dim).
        """

        enc_out = src  # shape: (batch_size, padded_src_len, hidden_dim)

        # 1. Extend src_key_padding_mask to (batch_size, 1, padded_src_len, padded_src_len) to be compatible with MHA.
        # 1 in the above shape will be expanded to number of heads in MHA.
        assert src_key_padding_mask.dtype == torch.bool
        src_key_padding_mask = src_key_padding_mask.unsqueeze(1).unsqueeze(
            1
        )  # shape: (batch_size, 1, 1, padded_src_len)
        src_key_padding_mask = src_key_padding_mask.expand(
            -1, -1, src_key_padding_mask.size(3), -1
        )  # shape: (batch_size, 1, padded_src_len, padded_src_len)

        # 2. Loop through the encoder layers.
        for layer in self.transformer:
            enc_out = layer(
                enc_out, src_key_padding_mask
            )  # shape: (batch_size, padded_src_len, hidden_dim)

        # 3. Layer normalization before LM head.
        enc_out = self.layer_norm(enc_out)  # shape: (batch_size, padded_src_len, hidden_dim)

        return enc_out

    def update_dropout(
        self,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
    ):
        for layer in self.transformer:
            layer.update_dropout(dropout, attention_dropout)



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/translate/decode_strategy.py
"""
import random
from copy import deepcopy

import numpy as np
import torch


def tile(x: torch.Tensor, count: int, dim: int = 0):
    """Tiles x on dimension dim count times.

    Args:
        x (torch.Tensor): The tensor to tile.
        memory_bank as an example x, shape (padded_src_len, batch_size, hidden_dim), where dim=1.
        count (int): The number of times to tile x.
        dim (int): The dimension to tile x on. Default is 0.

    Returns:
        x (torch.Tensor): The tiled tensor.
        For memory_bank as an example x, shape (padded_src_len, batch_size * count/beam_size, hidden_dim), where dim=1.

    """
    perm = list(range(len(x.size())))
    if dim != 0:
        perm[0], perm[dim] = perm[dim], perm[0]
        x = x.permute(perm).contiguous()

    out_size = list(x.size())
    out_size[0] *= count
    batch = x.size(0)
    x = (
        x.view(batch, -1)
        .transpose(0, 1)
        .repeat(count, 1)
        .transpose(0, 1)
        .contiguous()
        .view(*out_size)
    )

    if dim != 0:
        x = x.permute(perm).contiguous()
    return x


def set_random_seed(seed, is_cuda):
    """Sets the random seed."""
    if seed > 0:
        torch.manual_seed(seed)
        # this one is needed for torchtext random call (shuffled iterator)
        # in multi gpu it ensures datasets are read in the same order
        random.seed(seed)
        # some cudnn methods can be random even after fixing the seed
        # unless you tell it to be deterministic
        torch.backends.cudnn.deterministic = True
        # This one is needed for various tranfroms
        np.random.seed(seed)

    if is_cuda and seed > 0:
        # These ensure same initialization in multi gpu mode
        torch.cuda.manual_seed(seed)


def report_matrix(row_label, column_label, matrix):
    header_format = "{:>10.10} " + "{:>10.7} " * len(row_label)
    row_format = "{:>10.10} " + "{:>10.7f} " * len(row_label)
    output = header_format.format("", *row_label) + "\n"
    for word, row in zip(column_label, matrix):
        max_index = row.index(max(row))
        row_format = row_format.replace("{:>10.7f} ", "{:*>10.7f} ", max_index + 1)
        row_format = row_format.replace("{:*>10.7f} ", "{:>10.7f} ", max_index)
        output += row_format.format(word, *row) + "\n"
        row_format = "{:>10.10} " + "{:>10.7f} " * len(row_label)
    return output


class DecodeStrategy(object):
    """Base class for generation strategies.

    Args:
        pad (int): Magic integer in output vocab.
        bos (int): Magic integer in output vocab.
        eos (int): Magic integer in output vocab.
        unk (int): Magic integer in output vocab.
        batch_size (int): Current batch size.
        parallel_paths (int): Decoding strategies like beam search
            use parallel paths. Each batch is repeated ``parallel_paths``
            times in relevant state tensors.
        min_length (int): Shortest acceptable generation, not counting
            begin-of-sentence or end-of-sentence.
        max_length (int): Longest acceptable sequence, not counting
            begin-of-sentence (presumably there has been no EOS
            yet if max_length is used as a cutoff).
        ban_unk_token (Boolean): Whether unk token is forbidden
        block_ngram_repeat (int): Block beams where
            ``block_ngram_repeat``-grams repeat.
        exclusion_tokens (set[int]): If a gram contains any of these
            tokens, it may repeat.
        return_attention (bool): Whether to work with attention too. If this
            is true, it is assumed that the decoder is attentional.

    Attributes:
        pad (int): See above.
        bos (int): See above.
        eos (int): See above.
        unk (int): See above.
        predictions (list[list[LongTensor]]): For each batch, holds a
            list of beam prediction sequences.
        scores (list[list[FloatTensor]]): For each batch, holds a
            list of scores.
        attention (list[list[FloatTensor or list[]]]): For each
            batch, holds a list of attention sequence tensors
            (or empty lists) having shape ``(step, inp_seq_len)`` where
            ``inp_seq_len`` is the length of the sample (not the max
            length of all inp seqs).
        alive_seq (LongTensor): Shape ``(B x parallel_paths, step)``.
            This sequence grows in the ``step`` axis on each call to
            :func:`advance()`.
        is_finished (ByteTensor or NoneType): Shape
            ``(B, parallel_paths)``. Initialized to ``None``.
        alive_attn (FloatTensor or NoneType): If tensor, shape is
            ``(step, B x parallel_paths, inp_seq_len)``, where ``inp_seq_len``
            is the (max) length of the input sequence.
        target_prefix (LongTensor or NoneType): If tensor, shape is
            ``(B x parallel_paths, prefix_seq_len)``, where ``prefix_seq_len``
            is the (max) length of the pre-fixed prediction.
        min_length (int): See above.
        max_length (int): See above.
        ban_unk_token (Boolean): See above.
        block_ngram_repeat (int): See above.
        exclusion_tokens (set[int]): See above.
        return_attention (bool): See above.
        done (bool): See above.
    """

    def __init__(
        self,
        pad: int = 1,
        bos: int = 2,
        eos: int = 3,
        unk: int = 0,
        batch_size: int = 32,
        parallel_paths: int = 10,  # beam size
        min_length: int = 0,
        max_length: int = 512,
        block_ngram_repeat: int = 0,
        exclusion_tokens: set[int] = set(),
        ban_unk_token: bool = False,
        return_attention: bool = False,
    ):
        # magic indices
        self.pad = pad
        self.bos = bos
        self.eos = eos
        self.unk = unk

        self.batch_size = batch_size
        self.parallel_paths = parallel_paths
        self.beam_size = parallel_paths

        self.min_length = min_length
        self.max_length = max_length
        self.block_ngram_repeat = block_ngram_repeat
        self.exclusion_tokens = exclusion_tokens
        self.ban_unk_token = ban_unk_token
        self.return_attention = return_attention

        # result caching
        self.predictions: list = [[] for _ in range(batch_size)]
        self.scores: list = [[] for _ in range(batch_size)]
        self.attention: list = [[] for _ in range(batch_size)]
        self.hypotheses: list = [[] for _ in range(batch_size)]

        self.alive_attn = None
        self.done = False

        n_paths = batch_size * parallel_paths
        self.forbidden_tokens: list = [dict() for _ in range(n_paths)]

    def get_device_from_memory_bank(self, memory_bank):
        if isinstance(memory_bank, tuple):
            mb_device = memory_bank[0].device
        else:
            mb_device = memory_bank.device
        return mb_device

    def initialize_tile(
        self, memory_bank: torch.Tensor, src_lengths, src_map=None, target_prefix=None
    ):
        """Tile everything for the beam search loop.

        Args:
            memory_bank (FloatTensor): The encoder memory bank. shape (padded_src_len, batch_size, hidden_dim).
            src_lengths (LongTensor): The source lengths. shape (batch_size,).
        """

        def fn_map_state(state, dim):
            return tile(state, self.beam_size, dim=dim)

        if isinstance(memory_bank, tuple):
            memory_bank = tuple(tile(x, self.beam_size, dim=1) for x in memory_bank)
        elif memory_bank is not None:
            memory_bank = tile(memory_bank, self.beam_size, dim=1)
        if src_map is not None:
            src_map = tile(src_map, self.beam_size, dim=1)

        self.memory_lengths = tile(src_lengths, self.beam_size)  # shape: (batch_size * beam_size,)
        if target_prefix is not None:
            target_prefix = tile(target_prefix, self.beam_size, dim=1)

        return fn_map_state, memory_bank, src_map, target_prefix

    def initialize(self, memory_bank, src_lengths, src_map=None, device=None, target_prefix=None):
        """DecodeStrategy subclasses should override :func:`initialize()`.

        Args:
            memory_bank (FloatTensor): The encoder memory bank. shape (padded_src_len, batch_size * beam_size, hidden_dim).
            src_lengths (LongTensor): The source lengths. shape (batch_size * beam_size,).

        `initialize` should be called before all actions.
        used to prepare necessary ingredients for decode.
        """
        if device is None:
            device = torch.device("cpu")
        self.alive_seq = torch.full(
            [self.batch_size * self.parallel_paths, 1], self.bos, dtype=torch.long, device=device
        )
        self.is_finished = torch.zeros(
            [self.batch_size, self.parallel_paths], dtype=torch.uint8, device=device
        )
        if target_prefix is not None:
            seq_len, batch_size, n_feats = target_prefix.size()
            assert (
                batch_size == self.batch_size * self.parallel_paths
            ), "forced target_prefix should've extend to same number of path!"
            target_prefix_words = target_prefix[:, :, 0].transpose(0, 1)
            target_prefix = target_prefix_words[:, 1:]  # remove bos

            # fix length constraint and remove eos from count
            prefix_non_pad = target_prefix.ne(self.pad).sum(dim=-1).tolist()
            self.max_length += max(prefix_non_pad) - 1
            self.min_length += min(prefix_non_pad) - 1

        self.target_prefix = target_prefix  # NOTE: forced prefix words
        return None, memory_bank, src_lengths, src_map

    def __len__(self):
        return self.alive_seq.shape[1]

    def ensure_min_length(self, log_probs):
        if len(self) <= self.min_length:
            log_probs[:, self.eos] = -1e20

    def ensure_unk_removed(self, log_probs):
        if self.ban_unk_token:
            log_probs[:, self.unk] = -1e20

    def ensure_max_length(self):
        # add one to account for BOS. Don't account for EOS because hitting
        # this implies it hasn't been found.
        if len(self) == self.max_length + 1:
            self.is_finished.fill_(1)

    def block_ngram_repeats(self, log_probs):
        """
        We prevent the beam from going in any direction that would repeat any
        ngram of size <block_ngram_repeat> more thant once.

        The way we do it: we maintain a list of all ngrams of size
        <block_ngram_repeat> that is updated each time the beam advances, and
        manually put any token that would lead to a repeated ngram to 0.

        This improves on the previous version's complexity:
           - previous version's complexity: batch_size * beam_size * len(self)
           - current version's complexity: batch_size * beam_size

        This improves on the previous version's accuracy;
           - Previous version blocks the whole beam, whereas here we only
            block specific tokens.
           - Before the translation would fail when all beams contained
            repeated ngrams. This is sure to never happen here.
        """

        # we don't block nothing if the user doesn't want it
        if self.block_ngram_repeat <= 0:
            return

        # we can't block nothing beam's too short
        if len(self) < self.block_ngram_repeat:
            return

        n = self.block_ngram_repeat - 1
        for path_idx in range(self.alive_seq.shape[0]):
            # we check paths one by one

            current_ngram = tuple(self.alive_seq[path_idx, -n:].tolist())
            forbidden_tokens = self.forbidden_tokens[path_idx].get(current_ngram, None)
            if forbidden_tokens is not None:
                log_probs[path_idx, list(forbidden_tokens)] = -10e20

    def maybe_update_forbidden_tokens(self):
        """We complete and reorder the list of forbidden_tokens"""

        # we don't forbid nothing if the user doesn't want it
        if self.block_ngram_repeat <= 0:
            return

        # we can't forbid nothing if beam's too short
        if len(self) < self.block_ngram_repeat:
            return

        n = self.block_ngram_repeat

        forbidden_tokens = list()
        for path_idx, seq in zip(self.select_indices, self.alive_seq):

            # Reordering forbidden_tokens following beam selection
            # We rebuild a dict to ensure we get the value and not the pointer
            forbidden_tokens.append(deepcopy(self.forbidden_tokens[path_idx]))

            # Grabing the newly selected tokens and associated ngram
            current_ngram = tuple(seq[-n:].tolist())

            # skip the blocking if any token in current_ngram is excluded
            if set(current_ngram) & self.exclusion_tokens:
                continue

            forbidden_tokens[-1].setdefault(current_ngram[:-1], set())
            forbidden_tokens[-1][current_ngram[:-1]].add(current_ngram[-1])

        self.forbidden_tokens = forbidden_tokens

    def target_prefixing(self, log_probs):
        """Fix the first part of predictions with `self.target_prefix`.

        Args:
            log_probs (FloatTensor): logits of size ``(B, vocab_size)``.

        Returns:
            log_probs (FloatTensor): modified logits in ``(B, vocab_size)``.
        """
        _B, vocab_size = log_probs.size()
        step = len(self)
        if self.target_prefix is not None and step <= self.target_prefix.size(1):
            pick_idx = self.target_prefix[:, step - 1].tolist()  # (B)
            pick_coo = [
                [path_i, pick]
                for path_i, pick in enumerate(pick_idx)
                if pick not in [self.eos, self.pad]
            ]
            mask_pathid = [
                path_i for path_i, pick in enumerate(pick_idx) if pick in [self.eos, self.pad]
            ]
            if len(pick_coo) > 0:
                pick_coo = torch.tensor(pick_coo).to(self.target_prefix)
                pick_fill_value = torch.ones([pick_coo.size(0)], dtype=log_probs.dtype)
                # pickups: Tensor where specified index were set to 1, others 0
                pickups = torch.sparse_coo_tensor(
                    pick_coo.t(), pick_fill_value, size=log_probs.size(), device=log_probs.device
                ).to_dense()
                # dropdowns: opposite of pickups, 1 for those shouldn't pick
                dropdowns = torch.ones_like(pickups) - pickups
                if len(mask_pathid) > 0:
                    path_mask = torch.zeros(_B).to(self.target_prefix)
                    path_mask[mask_pathid] = 1
                    path_mask = path_mask.unsqueeze(1).to(dtype=bool)
                    dropdowns = dropdowns.masked_fill(path_mask, 0)
                # Minus dropdowns to log_probs making probabilities of
                # unspecified index close to 0
                log_probs -= 10000 * dropdowns
        return log_probs

    def maybe_update_target_prefix(self, select_index):
        """We update / reorder `target_prefix` for alive path."""
        if self.target_prefix is None:
            return
        # prediction step have surpass length of given target_prefix,
        # no need to further change this attr
        if len(self) > self.target_prefix.size(1):
            return
        self.target_prefix = self.target_prefix.index_select(0, select_index)

    def advance(self, log_probs, attn):
        """DecodeStrategy subclasses should override :func:`advance()`.

        Advance is used to update ``self.alive_seq``, ``self.is_finished``,
        and, when appropriate, ``self.alive_attn``.
        """

        raise NotImplementedError()

    def update_finished(self):
        """DecodeStrategy subclasses should override :func:`update_finished()`.

        ``update_finished`` is used to update ``self.predictions``,
        ``self.scores``, and other "output" attributes.
        """

        raise NotImplementedError()

    @property
    def current_predictions(self):
        raise NotImplementedError()



from retrochimera.inference.retrochimera import RetroChimeraModel
from retrochimera.inference.smiles_transformer import SmilesTransformerModel
from retrochimera.inference.template_classification import TemplateClassificationModel
from retrochimera.inference.template_localization import TemplateLocalizationModel


class BasicTemplateClassificationModel(TemplateClassificationModel):
    pass


class RetroChimeraDeNovoModel(SmilesTransformerModel):
    pass


class RetroChimeraEditModel(TemplateLocalizationModel):
    pass


__all__ = [
    "BasicTemplateClassificationModel",
    "RetroChimeraDeNovoModel",
    "RetroChimeraEditModel",
    "RetroChimeraModel",
    "SmilesTransformerModel",
    "TemplateClassificationModel",
    "TemplateLocalizationModel",
]



from abc import ABCMeta, abstractmethod
from typing import Callable, Generic, Iterable, TypeVar

import pytorch_lightning as pl
import torch
import torchmetrics
from pytorch_lightning.utilities import grad_norm
from torch.nn import functional as f

from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)

SampleType = TypeVar("SampleType")
ProcessedSampleType = TypeVar("ProcessedSampleType")
BatchType = TypeVar("BatchType")


class AbstractLightningModel(pl.LightningModule, metaclass=ABCMeta):
    """Base class for models based on PyTorch Lightning."""

    def __init__(
        self,
        learning_rate: float = 1e-3,
        learning_rate_decay_step_size: int = 100,
        learning_rate_decay_rate: float = 0.1,
        optimizer_betas: tuple[float, float] = (0.9, 0.999),
    ) -> None:
        super().__init__()

        self.save_hyperparameters(ignore=self.hyperparameters_excluded_from_checkpoint)

        self.learning_rate = learning_rate
        self.learning_rate_decay_step_size = learning_rate_decay_step_size
        self.learning_rate_decay_rate = learning_rate_decay_rate
        self.optimizer_betas = optimizer_betas

    @property
    def hyperparameters_excluded_from_checkpoint(self) -> list[str]:
        return []

    def log(self, name, value, *args, **kwargs) -> None:
        if name.startswith("val_") or name.startswith("test_"):
            # Sync validation and test metrics across devices.
            kwargs["sync_dist"] = True

        super().log(name, value.to(self.device), *args, **kwargs)

    @abstractmethod
    def ttv_step(self, batch, step_name):
        pass

    def training_step(self, batch, batch_idx):
        return self.ttv_step(batch, "train")

    def validation_step(self, batch, batch_idx):
        return self.ttv_step(batch, "val")

    def test_step(self, batch, batch_idx):
        return self.ttv_step(batch, "test")

    def setup(self, *args, **kwargs):
        # If running multi-GPU, make sure different GPUs have different seeds.
        if torch.distributed.is_initialized():
            initial_seed = torch.random.initial_seed()
            rank = torch.distributed.get_rank()
            new_seed = initial_seed + rank

            logger.info(f"Shifting random seed for GPU {rank} from {initial_seed} to {new_seed}")
            torch.manual_seed(new_seed)

        super().setup(*args, **kwargs)

    def on_before_optimizer_step(self, optimizer):
        self.log_dict(grad_norm(self, norm_type=2))

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.learning_rate, betas=self.optimizer_betas
        )
        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer,
            step_size=self.learning_rate_decay_step_size,
            gamma=self.learning_rate_decay_rate,
        )

        return {"optimizer": optimizer, "lr_scheduler": lr_scheduler}


class AbstractModel(
    AbstractLightningModel, Generic[SampleType, ProcessedSampleType, BatchType], metaclass=ABCMeta
):
    """Base class for models that are trainable via `cli/train.py`."""

    @abstractmethod
    def preprocess(
        self, samples: Iterable[SampleType], num_processes: int = 0
    ) -> Iterable[ProcessedSampleType]:
        pass

    @abstractmethod
    def collate(self, samples: list[ProcessedSampleType]) -> BatchType:
        pass


def log_classification_metrics(
    log_fn: Callable,
    step_name: str,
    batch_preds: torch.Tensor,
    batch_targets: torch.Tensor,
    batch_size: int,
) -> None:
    log_fn(
        f"{step_name}_acc",
        torchmetrics.functional.accuracy(preds=batch_preds, target=batch_targets),
        batch_size=batch_size,
    )

    for k in [5, 10, 50]:
        if k < batch_preds.shape[-1]:
            log_fn(
                f"{step_name}_top_{k}_acc",
                torchmetrics.functional.accuracy(preds=batch_preds, target=batch_targets, top_k=k),
                batch_size=batch_size,
            )

    if step_name != "train":
        num_classes = batch_preds.shape[-1]

        # The `retrieval_reciprocal_rank` function is unbatched and somewhat slow. We call it on
        # each sample separately, but only during validation/testing.
        reciprocal_ranks = [
            torchmetrics.functional.retrieval_reciprocal_rank(preds=preds, target=target)
            for preds, target in zip(batch_preds, f.one_hot(batch_targets, num_classes=num_classes))
        ]
        log_fn(
            f"{step_name}_mrr",
            torch.mean(torch.as_tensor(reciprocal_ranks)),
            batch_size=batch_size,
        )



"""Code adapted from RootAligned (https://github.com/otori-bird/retrosynthesis/blob/main/score.py).

Here it is modified to process predictions given as a tuple of (SMILES, probability).
"""

from rdkit import Chem


def canonicalize_smiles_clear_map(prediction, return_max_frag=True):
    """Convert a prediction tuple (SMILES, probability) into a canonicalized SMILES string.

    Args:
        prediction: Input tuple.
        return_max_frag: Whether to only return the SMILES of the maximum connected fragment.

    Returns:
        Canonicalized SMILES string, and (if `return_max_frag` is set) also the SMILES of the
        maximum connected fragment together with the probability.
    """
    smiles, probability = prediction
    mol = Chem.MolFromSmiles(smiles, sanitize=True)
    if mol is not None:
        [
            atom.ClearProp("molAtomMapNumber")
            for atom in mol.GetAtoms()
            if atom.HasProp("molAtomMapNumber")
        ]
        try:
            smi = Chem.MolToSmiles(mol, isomericSmiles=True)
        except Exception:
            if return_max_frag:
                return "", "", probability
            else:
                return ""
        if return_max_frag:
            sub_smi = smi.split(".")
            sub_mol = [Chem.MolFromSmiles(smiles, sanitize=True) for smiles in sub_smi]
            sub_mol_size = [
                (sub_smi[i], len(m.GetAtoms())) for i, m in enumerate(sub_mol) if m is not None
            ]
            if len(sub_mol_size) > 0:
                return (
                    smi,
                    canonicalize_smiles_clear_map(
                        (sorted(sub_mol_size, key=lambda x: x[1], reverse=True)[0][0], probability),
                        return_max_frag=False,
                    ),
                    probability,
                )
            else:
                return smi, "", probability
        else:
            return smi
    else:
        if return_max_frag:
            return "", "", probability
        else:
            return ""


def compute_rank(
    prediction: list[list[tuple[str, str, float]]], alpha: float = 1.0, opt=None
) -> tuple[dict[tuple[str, str], float], dict[tuple[str, str], list[tuple[int, float]]], int, int]:
    """Compute the rank of each prediction.

    Args:
        prediction: List of tuples (SMILES, max_frag_smiles, probability) for each augmentation.
        alpha: Scaling factor for rank calculation.

    Returns:
        A tuple of:
        - rank: Dictionary mapping (SMILES, max_frag_smiles) to scores.
        - position_prob_info: Dictionary mapping (SMILES, max_frag_smiles) to their relative rank
            (and probability) within each input augmentation's predictions.
        - adaptive_augmentation_size: Number of augmentations performed on the input molecule.
        - effective_augmentation_size: Number of augmentations where the model returned at least one
            valid prediction.
    """
    valid_score = [[k for k in range(len(prediction[j]))] for j in range(len(prediction))]
    rank: dict[tuple[str, str], float] = {}
    highest: dict[tuple[str, str], int] = {}
    position_prob_info: dict[tuple[str, str], list[tuple[int, float]]] = {}
    effective_augmentation_size = 0

    for j in range(len(prediction)):
        reaction_max_probability_map = {}
        for k in range(len(prediction[j])):
            if prediction[j][k][0] == "":
                valid_score[j][k] = opt.beam_size + 1
            if prediction[j][k][0:2] not in reaction_max_probability_map:
                reaction_max_probability_map[prediction[j][k][0:2]] = prediction[j][k][2]

        # error detection and deduplication
        de_error = [
            i[0][0:2]
            for i in sorted(list(zip(prediction[j], valid_score[j])), key=lambda x: x[1])
            if i[0][0] != ""
        ]
        cleaned_prediction = list(
            set(de_error)
        )  # cleaned means deduplicated, invalid smiles removed, sorted by probability
        cleaned_prediction.sort(key=de_error.index)
        if len(cleaned_prediction) != 0:
            effective_augmentation_size += 1
        for k, data in enumerate(cleaned_prediction):
            if data in rank:
                rank[data] += 1 / (alpha * k + 1)
                position_prob_info[data].append(
                    (k + 1, reaction_max_probability_map[data])
                )  # relative rank starting from 1, instead of 0
            else:
                rank[data] = 1 / (alpha * k + 1)
                position_prob_info[data] = [(k + 1, reaction_max_probability_map[data])]
            if data in highest:
                highest[data] = min(k, highest[data])
            else:
                highest[data] = k

    for key in rank.keys():
        rank[key] += highest[key] * -1e8

    adaptive_augmentation_size = len(prediction)

    return rank, position_prob_info, adaptive_augmentation_size, effective_augmentation_size



from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Iterable, Union

from syntheseus.reaction_prediction.utils.misc import parallelize

from retrochimera.data.preprocessing.template_classification import get_unique_template_id
from retrochimera.data.smiles_reaction_sample import SmilesReactionSample
from retrochimera.data.smiles_tokenizer import Tokenizer


@dataclass
class ProcessedSample:
    encoded_products: list[int]  # For pistachio and uspto50k, samples only have one main product
    encoded_reactants: list[int]
    template_id: int  # ID of the ground-truth template


def _process_sample(
    sample: tuple[str, str, int],
    input_preprocess_fn,
    output_preprocess_fn,
) -> ProcessedSample:
    products_smi, reactants_smi, template_id = sample
    encoded_products, _ = input_preprocess_fn([products_smi])
    encoded_reactants, _ = output_preprocess_fn([reactants_smi])
    return ProcessedSample(
        encoded_products=encoded_products[0],
        encoded_reactants=encoded_reactants[0],
        template_id=template_id,
    )


def preprocess_samples(
    samples: Iterable[SmilesReactionSample],
    rulebase_dir: Union[str, Path],
    tokenizer: Tokenizer,
    num_processes: int,
) -> Iterable[ProcessedSample]:
    # Extract (product smiles, reactants smiles, template id) from each sample.
    samples_processed = (
        (sample.raw_products_smiles, sample.raw_reactants_smiles, get_unique_template_id(sample))
        for sample in samples
    )

    # Preprocess the input and output smiles strings.
    input_preprocess_fn = partial(
        tokenizer.encode, pad=False, right_padding=True, add_begin_token=False, add_end_token=False
    )
    output_preprocess_fn = partial(
        tokenizer.encode, pad=False, right_padding=True, add_begin_token=True, add_end_token=True
    )

    yield from parallelize(
        partial(
            _process_sample,
            input_preprocess_fn=input_preprocess_fn,
            output_preprocess_fn=output_preprocess_fn,
        ),
        samples_processed,
        num_processes=num_processes,
    )



"""Functions to fix broken molecules, in particular heterocycles.

Code was adapted from RDKit cookbook:
https://rdkit.readthedocs.io/en/latest/Cookbook.html#cleaning-up-heterocycles
"""

from typing import Optional

from rdkit import Chem, RDLogger

from retrochimera.utils.logging import get_logger

lg = RDLogger.logger()
lg.setLevel(RDLogger.CRITICAL)

logger = get_logger(__name__)


def reverse_rxn(smi):
    r, a, p = smi.split(">")
    return f"{p}>{a}>{r}"


def frag_indices_to_mol(o_mol, indices):
    em = Chem.EditableMol(Chem.Mol())

    newIndices = {}
    for i, idx in enumerate(indices):
        em.AddAtom(o_mol.GetAtomWithIdx(idx))
        newIndices[idx] = i

    for i, idx in enumerate(indices):
        at = o_mol.GetAtomWithIdx(idx)
        for bond in at.GetBonds():
            if bond.GetBeginAtomIdx() == idx:
                oidx = bond.GetEndAtomIdx()
            else:
                oidx = bond.GetBeginAtomIdx()
            # make sure every bond only gets added once:
            if oidx < idx:
                continue
            em.AddBond(newIndices[idx], newIndices[oidx], bond.GetBondType())
    res = em.GetMol()
    res.ClearComputedProps()
    Chem.GetSymmSSSR(res)
    res.UpdatePropertyCache(False)
    res._idxMap = newIndices
    return res


def _recursively_modify_Ns(mol, matches, indices=None):
    if indices is None:
        indices = []
    res = None
    while len(matches) and res is None:
        tIndices = indices[:]
        nextIdx = matches.pop(0)
        tIndices.append(nextIdx)
        nm = Chem.Mol(mol.ToBinary())
        nm.GetAtomWithIdx(nextIdx).SetNoImplicit(True)
        nm.GetAtomWithIdx(nextIdx).SetNumExplicitHs(1)
        cp = Chem.Mol(nm.ToBinary())
        try:
            Chem.SanitizeMol(cp)
        except ValueError:
            res, indices = _recursively_modify_Ns(nm, matches, indices=tIndices)
        else:
            indices = tIndices
            res = cp
    return res, indices


def adjust_aromatic_Ns(m: Chem.Mol, nitrogenPattern: str = "[n&D2&H0;r5,r6]") -> Optional[Chem.Mol]:
    """Fix issues with aromatic nitrogens in rings.

    Default pattern matches Ns in 5-rings and 6-rings in order to be able to fix `O=c1ccncc1`.
    """
    Chem.GetSymmSSSR(m)
    m.UpdatePropertyCache(False)

    # break non-ring bonds linking rings:
    em = Chem.EditableMol(m)
    linkers = m.GetSubstructMatches(Chem.MolFromSmarts("[r]!@[r]"))
    plsFix = set()
    for a, b in linkers:
        em.RemoveBond(a, b)
        plsFix.add(a)
        plsFix.add(b)
    nm = em.GetMol()
    for at in plsFix:
        at = nm.GetAtomWithIdx(at)
        if at.GetIsAromatic() and at.GetAtomicNum() == 7:
            at.SetNumExplicitHs(1)
            at.SetNoImplicit(True)

    # build molecules from the fragments:
    fragLists = Chem.GetMolFrags(nm)
    frags = [frag_indices_to_mol(nm, x) for x in fragLists]

    # loop through the fragments in turn and try to aromatize them:
    for frag in frags:
        cp = Chem.Mol(frag.ToBinary())
        try:
            Chem.SanitizeMol(cp)
        except ValueError:
            matches = [x[0] for x in frag.GetSubstructMatches(Chem.MolFromSmarts(nitrogenPattern))]
            lres, indices = _recursively_modify_Ns(frag, matches)
            if not lres:
                return None
            else:
                revMap = {}
                for k, v in frag._idxMap.items():
                    revMap[v] = k
                for idx in indices:
                    oatom = m.GetAtomWithIdx(revMap[idx])
                    oatom.SetNoImplicit(True)
                    oatom.SetNumExplicitHs(1)

    return m


def fix_mol(mol: Chem.Mol) -> Optional[str]:
    """Fix weird heterocycles.

    Args:
        mol: RDKit molecule to fix which may be in a broken state.

    Returns:
        SMILES of the fixed molecule, or `None` if it could not have been fixed.
    """
    try:
        mol.UpdatePropertyCache(False)
        cp = Chem.Mol(mol.ToBinary())
        Chem.SanitizeMol(cp)
        mol = cp
        return Chem.MolToSmiles(mol)
    except ValueError:
        logger.debug(f"fix_mol: {Chem.MolToSmiles(mol)}")

        mol_adjusted = adjust_aromatic_Ns(mol)
        if mol_adjusted is None:
            return None

        try:
            Chem.SanitizeMol(mol_adjusted)
            return Chem.MolToSmiles(mol_adjusted)
        except ValueError:
            return None



import dataclasses
from abc import abstractmethod
from collections import Counter, defaultdict
from pathlib import Path
from typing import Any, Iterable, Optional

from rdkit import Chem
from syntheseus import Bag, Molecule
from syntheseus.interface.molecule import SMILES_SEPARATOR
from syntheseus.interface.reaction import REACTION_SEPARATOR
from syntheseus.reaction_prediction.chem.utils import ATOM_MAPPING_PROP_NAME, remove_atom_mapping
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample
from syntheseus.reaction_prediction.utils.misc import parallelize
from tqdm import tqdm

from retrochimera.chem.utils import get_all_atom_mapping_nums, get_mapped_atoms
from retrochimera.utils.logging import get_logger

logger = get_logger(__name__)


def parse_sample(sample: tuple[str, dict[str, Any]]) -> Optional[ReactionSample]:
    """Parse a single sample (reaction SMILES, with its metadata)."""
    reaction_smiles, metadata = sample

    return ReactionSample.from_reaction_smiles(
        reaction_smiles=reaction_smiles, mapped=True, metadata=metadata
    )


def parse_samples(
    samples: list[str], num_processes: int, source: Optional[str] = None
) -> list[ReactionSample]:
    """Parse a list of samples (input strings) in parallel.

    Args:
        samples: List of input strings to parse.
        num_processes: Number of parallel processes to use.
        source: An optional tag to include into the metadata.

    Returns:
        List of samples (`ReactionSample`), excluding those that could not be parsed. Each parsed
        sample includes metadata such as the original string, `source`, and its index in `samples`.
    """
    # Parse samples in parallel; turn off chunking as we store all outputs in memory anyway. Here,
    # we first extract reaction SMILES from input strings (handling Pistachio-style format that
    # contains extra info); this is a no-op for other formats that only have the reaction SMILES.
    results = tqdm(
        parallelize(
            parse_sample,
            (
                (
                    sample.split("\t")[0].split(" ")[0],
                    {"original_str": sample, "source": source, "idx_in_source": idx},
                )
                for idx, sample in enumerate(samples)
            ),
            num_processes=num_processes,
            num_chunks_per_process_per_segment=None,
        ),
        total=len(samples),
    )
    return [result for result in results if result is not None]


class ProcessingStep:
    def __init__(self, name: str, output_dir: Path) -> None:
        self._name = name
        self._output_dir = output_dir
        self._logs_pending: str = ""
        self._samples_saved: dict[str, list[str]] = defaultdict(list)
        self._sample_counts: Counter[str] = Counter()

    def save_sample(
        self, sample: ReactionSample, prefix: str = "filtered", suffix: str = ""
    ) -> None:
        self._samples_saved[f"{prefix}_by_step_{self._name}{suffix}"].append(
            sample.metadata["original_str"]  # type: ignore[typeddict-item]
        )

    def process_samples(self, samples: Iterable[ReactionSample]) -> Iterable[ReactionSample]:
        num_samples = 0
        for sample in samples:
            processed_sample = self(sample)
            if processed_sample is not None:
                num_samples += 1
                yield processed_sample

        self._logs_pending += f"\n{'=' * 30} Step {self._name} {'=' * 30}\n"

        for key, count in self._sample_counts.items():
            self._logs_pending += f"Counted {count} samples {key}\n"

        for key, original_strings in self._samples_saved.items():
            filename = f"{key}.txt"
            self._logs_pending += f"Saving {len(original_strings)} samples under {filename}\n"

            with open(self._output_dir / filename, "w") as f:
                for original_string in original_strings:
                    f.write(str(original_string) + "\n")

        self.flush_additional_data()
        self._logs_pending += f"Left with {num_samples} samples\n"

        logger.info(self._logs_pending)

        self._logs_pending = ""
        self._samples_saved = defaultdict(list)
        self._sample_counts = Counter()

    def flush_additional_data(self) -> None:
        """Flush any extra stats or metadata after processing a stream of samples."""
        pass

    @abstractmethod
    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        """Process a sample and return its (repaired) version or `None` if should be discarded."""
        pass


def replace_reaction_smiles(sample: ReactionSample, new_reaction_smiles: str) -> ReactionSample:
    assert sample.mapped_reaction_smiles is not None

    old_str: str = sample.metadata.get("repaired_str", sample.metadata["original_str"])  # type: ignore
    new_str = old_str.replace(sample.mapped_reaction_smiles, new_reaction_smiles)

    sample = dataclasses.replace(sample, mapped_reaction_smiles=new_reaction_smiles)
    sample.metadata["repaired_str"] = new_str  # type: ignore[typeddict-item]

    return sample


class NumReactantsProcessingStep(ProcessingStep):
    """Removes samples with too many reactants."""

    def __init__(self, max_reactants_num: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._max_num_reactants = max_reactants_num

    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        if len(sample.reactants) > self._max_num_reactants:
            self.save_sample(sample)
            return None

        return sample


class OneMainProductProcessingStep(ProcessingStep):
    """Tries to select main product for each sample and removes if not possible."""

    def __init__(
        self, min_product_atoms: int, common_products: set[Molecule], *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self._min_product_atoms = min_product_atoms
        self._common_products = common_products

    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        main_products = []
        for product in sample.products:
            if (
                product not in self._common_products
                and product.rdkit_mol.GetNumAtoms() >= self._min_product_atoms
            ):
                main_products.append(product)

        assert len(sample.products) > 0
        if len(sample.products) == 1:
            self._sample_counts["originally with 1 product"] += 1
        elif len(sample.products) == 2:
            self._sample_counts["originally with 2 products"] += 1
        else:
            self._sample_counts["originally with 3 or more products"] += 1

        if len(main_products) != 1:  # keep samples where there is exactly one main product
            if len(sample.products) == 1:
                self.save_sample(sample, suffix="_originally_with_1_product")
            elif len(sample.products) == 2:
                self.save_sample(sample, suffix="_originally_with_2_products")
            else:
                self.save_sample(sample, suffix="_originally_with_3_or_more_products")
            return None

        [main_product] = main_products

        assert sample.mapped_reaction_smiles is not None
        mapped_reactants, reagents, mapped_products = sample.mapped_reaction_smiles.split(
            REACTION_SEPARATOR
        )

        for mapped_product in mapped_products.split(SMILES_SEPARATOR):
            if Molecule(remove_atom_mapping(mapped_product)) == main_product:
                mapped_main_product = mapped_product
                break
        else:
            assert False, "Could not extract mapped main product SMILES"

        return replace_reaction_smiles(
            dataclasses.replace(sample, products=Bag([main_product])),
            new_reaction_smiles=REACTION_SEPARATOR.join(
                [mapped_reactants, reagents, mapped_main_product]
            ),
        )


class NumAtomsProcessingStep(ProcessingStep):
    """Removes samples with large product or large ratio between reactants and product."""

    def __init__(
        self, max_product_atoms: int, max_reactants_to_product_ratio: int, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self._max_product_atoms = max_product_atoms
        self._max_reactants_to_product_ratio = max_reactants_to_product_ratio

    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        reactant_atoms = sum(r.rdkit_mol.GetNumAtoms() for r in sample.reactants)
        product_atoms = sum(p.rdkit_mol.GetNumAtoms() for p in sample.products)

        if product_atoms > self._max_product_atoms:
            self.save_sample(sample, suffix="_product_too_large")
            return None

        if reactant_atoms / product_atoms > self._max_reactants_to_product_ratio:
            self.save_sample(sample, suffix="_reactant_to_product_ratio_too_large")
            return None

        return sample


class ProductAmongReactantsProcessingStep(ProcessingStep):
    """Removes samples where a product molecule is one of the reactants."""

    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        if set(sample.reactants) & set(sample.products):
            self.save_sample(sample)
            return None

        return sample


class AtomMappingProcessingStep(ProcessingStep):
    """Tries to fix atom mapping information and removes the sample if not possible."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._reactants_removed: set[str] = set()

    def flush_additional_data(self) -> None:
        self._logs_pending += f"Removed {len(self._reactants_removed)} unmapped reactant sets\n"

        with open(self._output_dir / f"reactants_removed_by_step_{self._name}.txt", "w") as f:
            for reactant in self._reactants_removed:
                f.write(reactant + "\n")

        self._reactants_removed = set()

    def __call__(self, sample: ReactionSample) -> Optional[ReactionSample]:
        assert sample.mapped_reaction_smiles is not None
        mapped_reactants, reagents, mapped_product = sample.mapped_reaction_smiles.split(
            REACTION_SEPARATOR
        )

        reactants = [
            Molecule(reactant_smiles, canonicalize=False)
            for reactant_smiles in mapped_reactants.split(SMILES_SEPARATOR)
        ]
        product = Molecule(mapped_product, canonicalize=False)

        # Create sets of atom mapping numbers for reactants and product.
        atom_mapping_numbers_in_raw_reactants = set()
        for reactant in reactants:
            atom_mapping_numbers_in_raw_reactants |= set(
                get_all_atom_mapping_nums(reactant.rdkit_mol)
            )

        atom_mapping_numbers_in_raw_main_product = set(get_all_atom_mapping_nums(product.rdkit_mol))

        # Refine samples by clearing the atom mapping numbers that appear on only one side.
        repaired_flag = False
        atom_mapping_numbers_in_each_repaired_reactant = []
        atom_mapping_numbers_in_repaired_reactants = set()
        num_mapped_atoms_in_repaired_reactants = 0
        for reactant in reactants:
            reactant_atom_mapping_numbers = set()
            for reactant_atom in get_mapped_atoms(reactant.rdkit_mol):
                if reactant_atom.GetAtomMapNum() not in atom_mapping_numbers_in_raw_main_product:
                    reactant_atom.ClearProp(ATOM_MAPPING_PROP_NAME)
                    repaired_flag = True
                else:
                    reactant_atom_mapping_numbers.add(reactant_atom.GetAtomMapNum())
                    atom_mapping_numbers_in_repaired_reactants.add(reactant_atom.GetAtomMapNum())
                    num_mapped_atoms_in_repaired_reactants += 1
            atom_mapping_numbers_in_each_repaired_reactant.append(reactant_atom_mapping_numbers)

        atom_mapping_numbers_in_repaired_main_product = set()
        num_mapped_atoms_in_repaired_main_product = 0
        for product_atom in get_mapped_atoms(product.rdkit_mol):
            if product_atom.GetAtomMapNum() not in atom_mapping_numbers_in_raw_reactants:
                product_atom.ClearProp(ATOM_MAPPING_PROP_NAME)
                repaired_flag = True
            else:
                atom_mapping_numbers_in_repaired_main_product.add(product_atom.GetAtomMapNum())
                num_mapped_atoms_in_repaired_main_product += 1

        if repaired_flag:
            self.save_sample(sample, prefix="repaired", suffix="_removed_lone_atom_mapping_numbers")

        # Drop samples with double mapped atoms in the main product.
        if (
            len(atom_mapping_numbers_in_repaired_main_product)
            != num_mapped_atoms_in_repaired_main_product
        ):
            self.save_sample(sample, suffix="_double_mapped_product")
            return None

        # Drop samples with double mapped atoms in the reactants.
        if (
            len(atom_mapping_numbers_in_repaired_reactants)
            != num_mapped_atoms_in_repaired_reactants
        ):
            self.save_sample(sample, suffix="_double_mapped_reactants")
            return None

        # Drop samples without atom mapping numbers.
        if (
            len(atom_mapping_numbers_in_repaired_main_product) == 0
            or len(atom_mapping_numbers_in_repaired_reactants) == 0
        ):
            self.save_sample(sample, suffix="_no_atom_mapping")
            return None

        # Refine samples by removing reactants that don't contribute atoms to the product.
        refined_mapped_reactants_smiles_list: list[str] = []
        removed_reactants: list[str] = []
        remaining_unmapped_reactant_mols: list[Molecule] = []
        for i, (reactant, unmapped_reactant_mol) in enumerate(zip(reactants, sample.reactants)):
            if len(
                atom_mapping_numbers_in_each_repaired_reactant[i]
                - atom_mapping_numbers_in_repaired_main_product
            ) < len(atom_mapping_numbers_in_each_repaired_reactant[i]):
                refined_mapped_reactants_smiles_list.append(Chem.MolToSmiles(reactant.rdkit_mol))
                remaining_unmapped_reactant_mols.append(unmapped_reactant_mol)
            else:
                removed_reactants.append(Chem.MolToSmiles(reactant.rdkit_mol))

        refined_mapped_reactants_smiles = SMILES_SEPARATOR.join(
            refined_mapped_reactants_smiles_list
        )

        if removed_reactants:
            self.save_sample(sample, prefix="repaired", suffix="_removed_unmapped_reactants")
            self._reactants_removed.add(SMILES_SEPARATOR.join(removed_reactants))

        return replace_reaction_smiles(
            dataclasses.replace(sample, reactants=Bag(remaining_unmapped_reactant_mols)),
            new_reaction_smiles=f"{refined_mapped_reactants_smiles}>{reagents}>{Chem.MolToSmiles(product.rdkit_mol)}",
        )



from functools import partial
from itertools import islice
from pathlib import Path
from typing import Any, Optional, Union

import h5py
import torch
from tqdm import tqdm

from retrochimera.data.dataset import DataFold, ReactionDataset
from retrochimera.models.lightning import AbstractModel
from retrochimera.models.smiles_transformer import SmilesTransformerModel
from retrochimera.utils.logging import get_logger
from retrochimera.utils.misc import save_into_h5py

logger = get_logger(__name__)


def preprocess_and_save(
    save_path: Union[str, Path],
    dataset: ReactionDataset,
    model: AbstractModel,
    num_processes: int,
    max_num_samples: Optional[int] = None,
) -> None:
    """Preprocesses raw data for any model type and saves as one HDF5 file per fold."""
    preprocess_fn = partial(model.preprocess, num_processes=num_processes)

    logger.info(f"Starting to save data under {save_path}")
    with h5py.File(save_path, mode="w", libver="latest") as file:
        for fold in DataFold:
            logger.info(f"Saving fold {fold.value}")

            raw_data_iterator = dataset[fold]
            num_samples = dataset.get_num_samples(fold)

            if max_num_samples is not None:
                logger.info(f"Truncating the fold to {max_num_samples} samples")

                raw_data_iterator = islice(raw_data_iterator, max_num_samples)
                num_samples = min(num_samples, max_num_samples)

            group = file.create_group(fold.value)
            for idx, data in enumerate(preprocess_fn(tqdm(raw_data_iterator, total=num_samples))):
                save_into_h5py(group, key=str(idx), data=data)

            if isinstance(model, SmilesTransformerModel):
                model.tokenizer.print_unknown_tokens()


def average_checkpoints(input_paths: list[Union[str, Path]], output_path: Union[str, Path]) -> None:
    """Compute an averaged checkpoint by uniformly averaging out a given list of checkpoints.

    Args:
        input_paths: Paths to the checkpoints to use for averaging.
        output_path: Path under which to save the final combined checkpoint.
    """
    combined_checkpoint: Optional[dict[str, Any]] = None
    num_combined = 0

    for path in input_paths:
        with open(path, "rb") as f:
            checkpoint = torch.load(f, map_location="cpu")

            if combined_checkpoint is None:
                combined_checkpoint = checkpoint
            else:
                for key, weights in checkpoint["state_dict"].items():
                    if key.endswith(".num_batches_tracked"):
                        # Batch counter in `BatchNorm` may differ slightly accross checkpoints but
                        # it's not needed during inference.
                        continue

                    if weights.dtype.is_floating_point:
                        combined_checkpoint["state_dict"][key] += weights
                    else:
                        if (combined_checkpoint["state_dict"][key] != weights).any():
                            raise ValueError(
                                f"Checkpoints to combine differ on key {key} "
                                f"which cannot be averaged due to having type {weights.dtype}"
                            )

            num_combined += 1

    assert combined_checkpoint is not None

    if num_combined > 1:
        for key, weights in combined_checkpoint["state_dict"].items():
            if weights.dtype.is_floating_point:
                weights /= float(num_combined)

    with open(output_path, "wb") as f:
        torch.save(combined_checkpoint, f)



from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence, Union

from rdkit import RDLogger
from syntheseus import Bag, Molecule, SingleProductReaction
from syntheseus.interface.molecule import molecule_bag_to_smiles
from syntheseus.interface.reaction import ReactionMetaData

from retrochimera.chem.rewrite import Rewrite, RewriteResult
from retrochimera.utils.logging import get_logger

lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)

logger = get_logger(__name__)


@dataclass
class ReactionRule:
    rxn: Rewrite
    smarts: str  # Reaction SMARTS
    rule_id: int  # ID for the rule (e.g. for template classification)
    rule_hash: Optional[str]  # Hash of the rule (useful for database lookup)
    name: Optional[str]  # Name of the rule, if available
    n_support: int = -1  # Number of times the rule appeared in the dataset it was extracted from


class RuleBase:
    DEFAULT_FILE_NAME = "template_lib.json"

    def __init__(self) -> None:
        self.rules: dict[int, ReactionRule] = {}
        self.hash_dict: dict[str, int] = {}

    def add_rule(
        self,
        smarts: str,
        rule_id: Optional[int] = None,
        rule_hash: Optional[str] = None,
        name: Optional[str] = None,
        n_support: int = -1,
    ) -> None:
        """Add a new rule to the rule base.

        Args:
            rxn_smarts: RDKit reaction smarts.
            rule_hash: Hash code of the rule.
            rule_id: ID of the rule, usually the label for Multi Class Classification.
            name: Name of the rule.
            n_support: Support of the rule i.e. number of times it is contained in the dataset it
                was extracted from.
        """

        if rule_id is None:
            rule_id = len(self.rules)

        rule = ReactionRule(
            rxn=Rewrite.from_rdkit(smarts),
            smarts=smarts,
            rule_id=rule_id,
            rule_hash=rule_hash,
            name=name,
            n_support=n_support,
        )

        self.rules[rule_id] = rule

        if rule_hash is not None:
            self.hash_dict[rule_hash] = rule_id

    def __contains__(self, item) -> bool:
        return item in self.rules

    def __getitem__(self, item) -> ReactionRule:
        return self.rules[item]

    def __len__(self) -> int:
        return len(self.rules)

    def get_rule(self, rule_id: int) -> Optional[ReactionRule]:
        return self.rules.get(rule_id)

    def get_rule_from_hash(self, hashcode: str) -> Optional[ReactionRule]:
        rule_id = self.hash_dict.get(hashcode)

        if rule_id is not None:
            return self.rules.get(rule_id)
        else:
            return None

    def save_to_file(self, dir: Union[str, Path], filename: str = DEFAULT_FILE_NAME) -> None:
        """Save all rules into a file using the JSONLines format."""
        output_dir = Path(dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        with open(output_dir / filename, mode="wt") as f:
            for rule in self.rules.values():
                rule_dict = {
                    key: value for (key, value) in dataclasses.asdict(rule).items() if key != "rxn"
                }
                f.write(json.dumps(rule_dict) + "\n")

    @staticmethod
    def load_from_file(
        dir: Union[str, Path],
        filename: str = DEFAULT_FILE_NAME,
        min_rule_support: Optional[int] = None,
        max_num_rules: Optional[int] = None,
        rule_ids: Optional[Iterable[int]] = None,
    ) -> RuleBase:
        """Load a `RuleBase` saved through `save_to_file`.

        Args:
            dir: Directory under which to look for the `RuleBase` file.
            filename: Expected name of the `RuleBase` file.
            min_rule_support: Minimum `n_support` for a rule to be kept.
            max_num_rules: Maximum number of lowest ID rules to be kept.
            rule_ids: Explicit IDs of rules to be kept.
        """
        rulebase = RuleBase()

        rule_ids_set: Optional[set[int]] = None
        if rule_ids is not None:
            rule_ids_set = set(rule_ids)

        with open(Path(dir) / filename) as f:
            lines = list(f)
            orig_size = len(lines)

            for line in lines:
                data = json.loads(line)

                if min_rule_support is not None and data["n_support"] < min_rule_support:
                    break

                if max_num_rules is not None and len(rulebase) == max_num_rules:
                    break

                if rule_ids_set is not None and data["rule_id"] not in rule_ids_set:
                    continue

                rulebase.add_rule(**data)

        logger.info(
            f"Loaded rulebase "
            f"(min_rule_support = {min_rule_support}, max_num_rules = {max_num_rules}); "
            f"original size: {orig_size}, final size: {len(rulebase)}"
        )

        return rulebase

    @staticmethod
    def load_rule_ids_from_file(
        dir: Union[str, Path], filename: str = DEFAULT_FILE_NAME
    ) -> list[int]:
        """Similar to `load_from_file`, but only loads the rule IDs, and hence is much faster."""
        return [json.loads(line)["rule_id"] for line in open(Path(dir) / filename)]


@dataclass
class RulePrediction:
    id: int  # ID of the rule that the model proposes to apply.
    prob: float  # Probability assigned by the model.

    # The `RulePrediction` object lives from the point when the rule is proposed in
    # `_predict_ranked_rules`, and until it is applied. Models may wish to retain some data computed
    # during the former to use during the latter; these are stored below.
    localization_scores: Optional[list[list[float]]] = None


class RuleBasedRetrosynthesizer:
    def __init__(
        self,
        *,
        num_templates_per_result: int = 10,
        min_num_rules_to_apply: int = 5,
        max_cumulative_prob: float = 1.0,
        apply_rules_timeout: Optional[float] = None,
        include_all_metadata: bool = False,
        **kwargs,
    ) -> None:
        """Initialize a rule-based model.

        Args:
            num_templates_per_result: Ratio of the maximum number of templates we should try
                applying to the number of results requested by the caller.
            min_num_rules_to_apply: Minimum number of rules to apply before potentially breaking due
                to accumulating a large total probability mass.
            max_cumulative_prob: If the total probability of the rules reaches this threshold, and
                there were at least `min_num_rules_to_apply` applied, then the application will be
                terminated, even if `rules_to_apply` has not been exhausted.
            apply_rules_timeout: Timeout passed to the rule application server; this is the maximum
                amount of time (in seconds) we will wait for the server to apply a *single* rule. If
                not set, we fallback to the default setting in the server.
            include_all_metadata: If set, model outputs will include detailed metadata (may require
                substantial disk space for saving the results).
        """
        super().__init__(**kwargs)  # In case this is one of several base classes in the MRO.

        self._server: Optional[Any] = None
        self._num_templates_per_result = num_templates_per_result
        self._min_num_rules_to_apply = min_num_rules_to_apply
        self._max_cumulative_prob = max_cumulative_prob
        self._apply_rules_timeout = apply_rules_timeout
        self._include_all_metadata = include_all_metadata

    def start_server(self, rulebase_dir: Union[str, Path]) -> None:
        """Instantiate the rule application server lazily."""

        # Local to avoid circular import.
        from retrochimera.chem.rule_application_server import RuleApplicationServer

        self._server = RuleApplicationServer(
            rulebase_dir=rulebase_dir, rule_application_kwargs=self._rule_application_server_kwargs
        )

    def _get_server(self):
        if self._server is None:
            raise RuntimeError("Class not ready as `start_server()` was not called")

        return self._server

    @property
    def _rule_application_server_kwargs(self) -> dict[str, Any]:
        return {}

    def predict(
        self, targets: list[Molecule], top_k: int = 50
    ) -> list[Sequence[SingleProductReaction]]:
        """Batched single step retrosynthesis prediction.

        Args:
            targets: Molecules to perform single step prediction on.
            top_k: Number of rules to apply. Each rule can potentially match in multiple locations,
                so number of predictions can be larger than `top_k` (this is intended behaviour).

        Returns:
            A list of sequences of `SingleProductReaction`s, where `i`-th entry corresponds the `i`-th
            input molecule.
        """
        return self.apply_top_rules(
            inputs=targets,
            batch_rules_to_apply=self._predict_ranked_rules(
                targets, top_k=self._num_templates_per_result * top_k
            ),
        )

    def _predict_ranked_rules(
        self, targets: list[Molecule], top_k=50
    ) -> list[list[RulePrediction]]:
        """Predict the `top_k` rule IDs to apply to the each molecule in the batch."""
        server = self._get_server()
        return [
            [
                RulePrediction(id=rule_id, prob=1.0 / len(server.rule_ids))
                for rule_id in server.rule_ids
            ]
            for _ in targets
        ]

    def apply_top_rules(
        self, inputs: list[Molecule], batch_rules_to_apply: list[list[RulePrediction]]
    ) -> list[Sequence[SingleProductReaction]]:
        """Apply the top rules provided.

        Args:
            inputs: Molecules to apply the rules to.
            batch_rules_to_apply: Rules to apply for each input (their IDs and probabilities), in
                the order they should be applied.

        Returns:
            A sequence of `SingleProductReaction`s containing several sets of predicted precursors for
            the input molecule.
        """
        # Truncate the rules to apply based on total probability.
        batch_rule_ids_to_apply: list[list[int]] = [[] for _ in inputs]
        for rules_to_apply, rule_ids_to_apply in zip(batch_rules_to_apply, batch_rule_ids_to_apply):
            total_prob = 0.0
            for rule_prediction in rules_to_apply:
                if (
                    len(rule_ids_to_apply) >= self._min_num_rules_to_apply
                    and total_prob >= self._max_cumulative_prob
                ):
                    break

                total_prob += rule_prediction.prob
                rule_ids_to_apply.append(rule_prediction.id)

        apply_rules_kwargs = (
            {"timeout": self._apply_rules_timeout} if self._apply_rules_timeout else {}
        )

        batch_raw_results = self._get_server().apply_rules(
            inputs=inputs, rule_ids_to_apply=batch_rule_ids_to_apply, **apply_rules_kwargs
        )

        batch_results: list[Sequence[SingleProductReaction]] = []
        for input, raw_results, rule_predictions in zip(
            inputs, batch_raw_results, batch_rules_to_apply
        ):
            results: list[RewriteResult] = []
            for raw_result, rule_prediction in zip(raw_results, rule_predictions):
                for r in raw_result:
                    r.metadata.update(
                        {
                            "reaction_id": rule_prediction.id,
                            "template_probability": rule_prediction.prob,
                        }
                    )

                    if rule_prediction.localization_scores is not None:
                        r.metadata["localization_scores"] = rule_prediction.localization_scores

                    results.append(r)

            self._rerank_results(results)

            if not self._include_all_metadata:
                # Some metadata cannot be removed earlier as it's needed for e.g. `_rerank_results`.
                keys_to_remove = ["localizations", "localization_scores"]
                for r in results:
                    for key in keys_to_remove:
                        if key in r.metadata:
                            del r.metadata[key]

            seen_outputs: set[Bag[Molecule]] = set()
            results_mols: list[Bag[Molecule]] = []
            results_metadata: list[ReactionMetaData] = []

            for r in results:
                if r.mols in seen_outputs:
                    continue

                seen_outputs.add(r.mols)

                metadata: ReactionMetaData = r.metadata  # type: ignore
                metadata["reaction_smiles"] = f"{molecule_bag_to_smiles(r.mols)}>>{input.smiles}"

                results_mols.append(r.mols)
                results_metadata.append(metadata)

            self._fill_prediction_probability(results_metadata)

            batch_results.append(
                [
                    SingleProductReaction(product=input, reactants=mols, metadata=metadata)
                    for mols, metadata in zip(results_mols, results_metadata)
                ]
            )

        return batch_results

    def _fill_prediction_probability(self, results_metadata: list[ReactionMetaData]) -> None:
        """Fill in the `probability` value for predictions for a single input."""
        for metadata in results_metadata:
            metadata["probability"] = metadata["template_probability"]  # type: ignore[typeddict-item]

    def _rerank_results(self, results: list[RewriteResult]) -> None:
        """Optionally reorder the predictions."""
        pass


def get_products(
    mol: Molecule,
    rule: Rewrite,
    cutoff: float = 0.2,
    return_localization: bool = False,
) -> list[RewriteResult]:
    """Applies the rule to the input molecule.

    Args:
        mol: Reactant bag as single `Molecule` object.
        rule: `Rewrite` object representing the rule to apply.
        cutoff: Only keep a result if its SMILES length is at least a `cutoff` fraction of the
            SMILES length of `mol`.
        return_localization: Whether to also extract the corresponding localization indices.

    Returns:
        A set of possible outcomes for applying this rule to the input mol.
    """

    result_smiles_min_length = len(mol.smiles) * cutoff
    results = rule.apply(mol, return_localization=return_localization)

    results_filtered = []
    for result in results:
        combined_smiles = molecule_bag_to_smiles(result.mols)
        if len(combined_smiles) > result_smiles_min_length:
            results_filtered.append(result)
        else:
            logger.debug(f"Result filtered out: {mol.smiles}>>{combined_smiles}")

    return results_filtered
